﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.ServiceModel;
using APlan.ViewModel;

namespace APlan.Service
{
    public enum AssignmentAction
    {
        Insert = 1, Update = 2, Delete = 3
    }

    /// <summary>
    /// the methods of this class are accessible by the clients via wcf
    /// </summary>
    [ServiceBehavior(ConcurrencyMode = ConcurrencyMode.Reentrant
        , InstanceContextMode = InstanceContextMode.Single)]    
    public class APlanService : IAPlanService
    {
        private Dictionary<string, IAPlanCallback> callbacks;
        private Dictionary<string, DateTime> lastAction;

        public APlanService()
        {
            callbacks = new Dictionary<string, IAPlanCallback>();
            lastAction = new Dictionary<string, DateTime>();
        }

        /// <summary>
        /// get all customers to the given query string
        /// </summary>
        /// <param name="text">search query</param>
        public void GetCustomer(string text)
        {
            ObservableCollection<CustIBaseView> ret = new ObservableCollection<CustIBaseView>();
            
            DataClassesDataContext context = new DataClassesDataContext();
            
            System.Data.Linq.DataLoadOptions dlo = new System.Data.Linq.DataLoadOptions();
            
            dlo.LoadWith<CustIBaseView>((CustIBaseView i) => i.Installations);
            dlo.LoadWith<CustIBaseView>((CustIBaseView i) => i.Assignments);
            context.LoadOptions = dlo;
            
            List<string> keywords = new List<string>(text.Split(' '));

            var cust = from c in context.CustIBaseViews                                  
                       select c;
            
            foreach(string word in keywords)
            {
                string help = word;
                cust = from c in cust 
                       where (c.Name.Contains(help) || c.City.Contains(help) || c.Nr.ToString().Contains(help)
                           || c.Street.Contains(help) || c.HNr.Contains(help) || c.Ibase.ToString().Contains(help)
                           || c.IBaseCity.Contains(help) || c.IBaseHNr.Contains(help) 
                           || c.IBaseStreet.Contains(help)) select c;                            
            }
            

            ret = new ObservableCollection<CustIBaseView>(cust.ToArray<CustIBaseView>());
            
            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            client.OnGetCustomerFinished(ret);

            string user = client.GetUserId();
            updateUserAction(user);
        }

        /// <summary>
        /// reads the customer to the given ibase key
        /// </summary>
        /// <param name="ibase"></param>
        public void GetCustomerByIBase(long ibase)
        {
            ObservableCollection<CustIBaseView> ret = new ObservableCollection<CustIBaseView>();
            
            DataClassesDataContext context = new DataClassesDataContext();

            var cust = from c in context.CustIBaseViews
                       where c.Ibase == ibase
                       select c;
            
            System.Data.Linq.DataLoadOptions dlo = new System.Data.Linq.DataLoadOptions();
            
            dlo.LoadWith<CustIBaseView>((CustIBaseView i) => i.Installations);
            dlo.LoadWith<CustIBaseView>((CustIBaseView i) => i.Assignments);
            context.LoadOptions = dlo;
            
            ret = new ObservableCollection<CustIBaseView>(cust.ToArray<CustIBaseView>());

            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            client.OnGetCustomerFinished(ret);

            string user = client.GetUserId();
            updateUserAction(user);
        }

        /// <summary>
        /// reads all assignments between the given dates
        /// </summary>
        /// <param name="fromd">date from</param>
        /// <param name="tod">date to</param>
        public void GetAssignments(DateTime fromd, DateTime tod)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            
            var assign = from assignm in context.AssignmentIBaseViews
                         where ( ( assignm.ToDate >= fromd )
                            && ( assignm.ToDate <= tod ) ) select assignm;

            
            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            ObservableCollection<AssignmentIBaseView> ret = new ObservableCollection<AssignmentIBaseView>(assign.ToArray<AssignmentIBaseView>());
                        
            client.OnGetAssignmentsFinished(ret);

            string user = client.GetUserId();
            updateUserAction(user);
         
        }

        /// <summary>
        /// set the update timestamp of the user
        /// </summary>
        /// <param name="user"></param>
        private void updateUserAction(string user)
        {
            if(lastAction.ContainsKey(user))
               lastAction[user] = DateTime.Now;
        }

        /// <summary>
        /// delivers all technicians
        /// </summary>
        public void GetEmployees()
        {
            DataClassesDataContext context = new DataClassesDataContext();
            System.Data.Linq.DataLoadOptions dlo = new System.Data.Linq.DataLoadOptions();
            
            dlo.LoadWith<Employee>((Employee e) => e.Cities);
            context.LoadOptions = dlo;
   
            var emp = from e in context.Employees
                      orderby e.Regiogroup
                      select e;

            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            ObservableCollection<Employee> ret = new ObservableCollection<Employee>(emp.ToArray<Employee>());

            var regio = from reg in context.Regiogroups
                        select reg;

            client.OnGetEmployeesFinished(ret, regio.ToArray<Regiogroup>());
        }

        /// <summary>
        /// reads the city to the given city id
        /// </summary>
        /// <param name="citynr">id of the city</param>
        /// <returns></returns>
        public City GetCity(long citynr)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            System.Data.Linq.DataLoadOptions dlo = new System.Data.Linq.DataLoadOptions();
            dlo.LoadWith<City>((City c) => c.Regiogroup1);
            context.LoadOptions = dlo;
            var city = (from c in context.Cities
                       where c.ID == citynr
                       select c).Single();

            return city;
        }
      
        /// <summary>
        /// updates the data of the ibase (phone and info)
        /// </summary>
        /// <param name="data"></param>
        public void UpdateIBase(CustIBaseView data)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            if (data != null)
            { 
                var ib = (IBase) ( from i in context.IBases
                           where i.Ibase1 == data.Ibase
                           select i ).Single();

                ib.Phone = data.Phone;
                ib.Info = data.Info;
            
                context.SubmitChanges();

                IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
                string user = client.GetUserId();
                updateUserAction(user);
            }
        }

        /// <summary>
        ///  make a suggestion for an new assignment
        /// </summary>
        /// <param name="cust">Customer</param>
        /// <param name="minDate">earliest date</param>
        /// <param name="maxDate">latest date</param>
        /// <param name="timeFrom">earliest time</param>
        /// <param name="timeTo">latest time</param0>
        public void CreateAssignment(CustIBaseView cust, DateTime minDate, DateTime maxDate, TimeSpan timeFrom, TimeSpan timeTo)
        { 
            DataClassesDataContext context = new DataClassesDataContext();
            Routeplan plan = new Routeplan(minDate, maxDate, timeFrom, timeTo);
            maxDate = maxDate.AddDays(1);
            var existingAssignments = from a in context.Assignments
                                      where  a.ToDate >= minDate
                                       && a.ToDate < maxDate 
                                       select a;

            plan.GenerateRoutes(existingAssignments);
            Assignment assignment = plan.CreateAssignment(cust);            
            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            string user = client.GetUserId();

            assignmentAction(AssignmentAction.Insert, user, DateTime.Now, assignment);
            IAPlanCallback callback = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            var assignmentIbaseView = (from av in context.AssignmentIBaseViews
                                      where av.Id == assignment.Id
                                      select av).Single();
            
            callback.OnCreateAssignmentCallback(assignmentIbaseView, true);
            
            assignmentAction(AssignmentAction.Insert, user, DateTime.Now, assignment);
            updateUserAction(user);
        }

        /// <summary>
        /// save assignment data for undo/redo
        /// </summary>
        /// <param name="act"></param>
        /// <param name="usr"></param>
        /// <param name="actiondate"></param>
        /// <param name="a"></param>
        private void assignmentAction(AssignmentAction act, string usr, DateTime actiondate, Assignment a)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            AssignmentUndoRedo ur = new AssignmentUndoRedo();
            ur.Action = (int)act;
            ur.ActionDate = actiondate;
            ur.ActionUser = usr;
            ur.Undo = true;
            ur.Emplyoee = a.Employee;
            ur.Fix = a.fix;
            ur.FromDate = a.FromDate;
            ur.IBase = a.IBase;
            ur.Id = a.Id;
            ur.State = a.State;
            ur.ToDate = a.ToDate;
            ur.Type = a.Type;
            context.AssignmentUndoRedos.InsertOnSubmit(ur);

            var redo = from r in context.AssignmentUndoRedos
                       where r.ActionUser == usr && r.Undo == false select r;
            context.AssignmentUndoRedos.DeleteAllOnSubmit(redo);
            callbacks[usr].UndoRedoAvailable(true, false);
            context.SubmitChanges();
        }

        /// <summary>
        /// create a routeplan
        /// </summary>
        /// <param name="cust">Customers</param>
        /// <param name="StartDate">start date</param>
        /// <param name="EndDate">end date</param>
        /// <param name="beginTime">begin time</param>
        /// <param name="endTime">end time</param>
        public void CreateRoutePlan(HeatMapViewModel[] cust, DateTime StartDate, DateTime EndDate, TimeSpan beginTime, TimeSpan endTime)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            List<CustIBaseView> openCust = new List<CustIBaseView>();
            int sumopencust = 0;
            //count the number of open customers
            foreach (HeatMapViewModel h in cust)
            {
                sumopencust += h.NumberOfCustomers;
            }
            DateTime toDate = StartDate.AddDays(sumopencust);
            Routeplan plan = new Routeplan(StartDate, toDate, beginTime, endTime);
            var existingAssignments = from a in context.Assignments
                                      where a.ToDate >= StartDate                                       
                                      select a;
            foreach(HeatMapViewModel h in cust)
            {
                var openCustdb = from c in context.CustIBaseViews
                               where c.nextMaint >= h.From
                                 && c.nextMaint <= h.To
                               select c;
                foreach (CustIBaseView c in openCustdb)
                {
                    int openState = (int)AssignmentState.open;
                    var openAssignments = from a in c.Assignments
                                          where a.State == openState
                                          select a;
                    if (openAssignments.Count() == 0)
                        openCust.Add(c);
                }
            }
            plan.GenerateRoutes(existingAssignments);
            plan.CreatePendingtours(openCust, StartDate);
            plan.DeleteEmptyRoutes();

            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();            
            plan.CalculateSavings();
            plan.FillGaps();
            client.OnRouteplanFinished();
            AssignmentDistr[] assignments = plan.GetAllAssignments();
            ObservableCollection<AssignmentIBaseView> ibview = new ObservableCollection<AssignmentIBaseView>();
            HeatMapViewModel[] hm = GetOpenCustomers();

            
            foreach (AssignmentDistr a in assignments)
            {
                AssignmentIBaseView ib = (from aib in context.AssignmentIBaseViews
                                          where aib.Id == a.DBAssignment.Id
                                          select aib).Single();
                foreach (IAPlanCallback c in callbacks.Values)
                {
                    c.OnCreateAssignmentCallback(ib, false);
                }
            }
            
            foreach (IAPlanCallback c in callbacks.Values)
            {
                c.OnHeatmapUpdate(hm);
            }

            
        }

        /// <summary>
        /// acknowledge routplan
        /// </summary>
        /// <param name="ack">true when accepted, otherwise decline</param>
        public void AcknowledgeRoutePlan(Boolean ack)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            var propAss = from a in context.Assignments
                          where a.State == (int)AssignmentState.proposal
                          select a;
            if (ack)
            {
                
                foreach(Assignment a in propAss)
                {
                    a.State = (int)AssignmentState.open;
                    
                }
                context.SubmitChanges();
                
            }
            else
            {
                context.Assignments.DeleteAllOnSubmit(propAss);
                context.SubmitChanges();
                
            }
            var assignments = from a in context.AssignmentIBaseViews
                              select a;
            ObservableCollection<AssignmentIBaseView> ret = new ObservableCollection<AssignmentIBaseView>(assignments.ToArray<AssignmentIBaseView>());
            foreach (IAPlanCallback c in callbacks.Values)
            {
                c.OnGetAssignmentsFinished(ret);
            }
        }

        /// <summary>
        /// remove given assignment
        /// </summary>
        /// <param name="a"></param>
        public void RemoveAssignment(Assignment a)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            var assign = (from av in context.Assignments
                          where av.Id == a.Id
                          select av).Single();
            context.Assignments.DeleteOnSubmit(assign);
            context.SubmitChanges();
            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            string user = client.GetUserId();
            assignmentAction(AssignmentAction.Delete,user, DateTime.Now, assign);
            updateUserAction(user);
        }

        /// <summary>
        /// subscribe user to the server
        /// </summary>
        /// <param name="usr">user id</param>
        public void Subscribe(string usr)
        {            
            IAPlanCallback callback = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            DataClassesDataContext context = new DataClassesDataContext();
            var empl = (from e in context.Employees
                       where e.User == usr
                       select e).Single();
            
            //todo: catch exception, when employee does´nt exist...
            
            if (!callbacks.ContainsKey(usr))
            {
                callbacks.Add(usr, callback);
                if(!lastAction.ContainsKey(usr))
                    lastAction.Add(usr, DateTime.Now);
                foreach (string u in callbacks.Keys)
                {
                    if (u != usr)
                    {
                        IAPlanCallback cb = callbacks[u];
                        cb.OnUserUpdated(usr, empl.name, 0);

                        //send the user data to the subscriber
                        string uname = (from e in context.Employees
                                        where e.User == u
                                        select e.name).Single();
                        int minutes = (lastAction[u] - System.DateTime.Now).Minutes;
                        callback.OnUserUpdated(u, uname, minutes);
                    }
                }
            }

            var msg = (from m in context.Messages orderby m.PostingDate descending select m).ToArray();
            callback.OnMessageUpdate(msg);

            setUndoRedoButtons(usr);
        }

        /// <summary>
        /// set the undo/redo buttons of the given user
        /// </summary>
        /// <param name="usr"></param>
        private void setUndoRedoButtons(string usr)
        {
            Boolean undo = false, redo = false;
            IAPlanCallback callback = callbacks[usr];
            DataClassesDataContext context = new DataClassesDataContext();
            var undos = from a in context.AssignmentUndoRedos
                        where a.ActionUser == usr
                         && a.Undo == true
                        select a;
            if (undos.Count() != 0)
                undo = true;
            var redos = from a in context.AssignmentUndoRedos
                        where a.ActionUser == usr
                         && a.Undo == false
                        select a;
            if (redos.Count() != 0)
                redo = true;
            callback.UndoRedoAvailable(undo, redo);
        }

        /// <summary>
        /// create an open failure
        /// </summary>
        /// <param name="ibase">id of the ibase</param>
        public void CreateOpenFailure(Int64 ibase)
        {
            Assignment a = new Assignment();
            a.IBase = ibase;
            a.FromDate = DateTime.MaxValue;
            a.ToDate = DateTime.MaxValue;
            a.Type = (short)AssignmentType.failure;
            a.State = (short)AssignmentState.open;
            a.fix = true;
            InsertAssignment(a);

            distributeOpenFailures();
        }
        /// <summary>
        /// remove the given open failure
        /// </summary>
        /// <param name="ibase">id of the ibase</param>
        public void RemoveOpenFailure(Int64 ibase)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            var assignment = (from a in context.Assignments
                              where a.IBase == ibase
                                && a.FromDate == DateTime.MaxValue
                              select a).Single();
            context.Assignments.DeleteOnSubmit(assignment);

            context.SubmitChanges();
        }

        /// <summary>
        /// distribute open failures to the clients
        /// </summary>
        private void distributeOpenFailures()
        {
            foreach (string u in callbacks.Keys)
            {
                IAPlanCallback cb = callbacks[u];
                cb.OnOpenFailureUpdate(readOpenFailures());
            }
        }

        /// <summary>
        /// read all open failures
        /// </summary>
        /// <returns>ObservableCollection with CustIBaseViews of open failures</returns>
        private ObservableCollection<CustIBaseView> readOpenFailures()
        {
            DataClassesDataContext context = new DataClassesDataContext();
            System.Data.Linq.DataLoadOptions dlo = new System.Data.Linq.DataLoadOptions();

            dlo.LoadWith<CustIBaseView>((CustIBaseView i) => i.Installations);
            dlo.LoadWith<CustIBaseView>((CustIBaseView i) => i.Assignments);
            context.LoadOptions = dlo;
            
            var failures = from f in context.Assignments
                           where f.FromDate == DateTime.MaxValue
                           select f;
            
            ObservableCollection<CustIBaseView> ret = new ObservableCollection<CustIBaseView>();

            foreach (Assignment f in failures)
            {
                var ib = (from i in context.CustIBaseViews
                          where i.Ibase == f.IBase
                          select i).Single();
                ret.Add(ib);
            }

            return ret;
        }

        public void GetOpenFailures()
        {
            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            ObservableCollection<CustIBaseView> ret = readOpenFailures();
            client.OnOpenFailureUpdate(ret);
        }

        /// <summary>
        /// unsubscribe the user
        /// </summary>
        /// <param name="usr"></param>
        public void Unsubscribe(string usr)
        {
            IAPlanCallback callback = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            callbacks.Remove(usr);
            foreach (string u in callbacks.Keys)
            {
                IAPlanCallback cb = callbacks[u];
                cb.OnUserRemoved(usr);
            }
        }

        /// <summary>
        /// get all subscribed users
        /// </summary>
        /// <returns>array of Employees</returns>
        public Employee[] GetUserList()
        {
            DataClassesDataContext context = new DataClassesDataContext();
            List<Employee> ret = new List<Employee>();
            foreach(String usr in callbacks.Keys)
            {
                Employee empl = (from e in context.Employees
                            where e.User == usr
                            select e).Single();
                ret.Add(empl);
            }
            return ret.ToArray();
        }

        
        /// <summary>
        /// get all open customers for HeatMap
        /// </summary>
        /// <returns></returns>
        public HeatMapViewModel[] GetOpenCustomers()
        {
            DataClassesDataContext context = new DataClassesDataContext();
            
            Dictionary<string, HeatMapViewModel> dic = new Dictionary<string, HeatMapViewModel>();
            var cust = from c in context.CustIBaseViews
                       where c.nextMaint != null
                         && c.Assignments.Count() == 0
                       orderby c.nextMaint
                       select c;

            foreach (CustIBaseView c in cust)
            {
                HeatMapViewModel h = new HeatMapViewModel();
                
                h.From = (DateTime)c.nextMaint;
                
                if (dic.ContainsKey(h.From.ToShortDateString()))
                {
                    dic[h.From.ToShortDateString()].NumberOfCustomers++;
                /*
                    custlist.Add(c);
                    dic[h.From.ToShortDateString()].OpenCustomers = custlist.ToArray();
                 * */
                }
                else
                {
                    h.NumberOfCustomers = 1;
                    dic.Add(h.From.ToShortDateString(), h);                    
                    /* custlist = new List<CustIBaseView>();
                    custlist.Add(c);*/
                }
            }

            return dic.Values.ToArray<HeatMapViewModel>();
        }

        /// <summary>
        /// read the details (customers) to the given heatmaps
        /// </summary>
        /// <param name="h"></param>
        /// <returns></returns>
        public HeatMapViewModel GetOpenCustomersDetail(HeatMapViewModel h)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            List<CustIBaseView> custlist = new List<CustIBaseView>();
            var cust = from c in context.CustIBaseViews
                       where c.nextMaint != null
                         && c.nextMaint >= h.From
                         && c.nextMaint <= h.To
                         && c.Assignments.Count() == 0
                       orderby c.nextMaint
                       select c;
            foreach (CustIBaseView c in cust)
            {
                custlist.Add(c);
            }
            h.OpenCustomers = custlist.ToArray();

            return h;
        }
        #region IScheduleService Members
        
        public void Test()
        {
            DataClassesDataContext context = new DataClassesDataContext();

            foreach(Customer c in context.Customers)
            {
                foreach (IBase i in c.IBases)
                {
                    if(c.nextMaint.CompareTo("#N/A")!=0)
                        i.nextMaint = DateTime.Parse(c.nextMaint);
                }
            }
            context.SubmitChanges();
        }
        public void InsertAssignment(Assignment a)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            AssignmentDistr ad = new AssignmentDistr();
            ad.DBAssignment = a;
            
            lock(context)
            {
                if (context.Assignments.Count() == 0)
                    a.Id = 1;
                else
                {
                    long maxid = (from assign in context.Assignments select assign.Id).Max();

                    a.Id = (long)maxid + 1;
                }
            }
            context.Assignments.InsertOnSubmit(a);
            context.SubmitChanges();
            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            string user = client.GetUserId();
            assignmentAction(AssignmentAction.Insert, user, DateTime.Now, a);

            updateUserAction(user);

            var ib = (from i in context.AssignmentIBaseViews
                          where i.Id == a.Id
                          select i).Single();
            ad.AssignIBaseView = ib;

            distributeOpenFailures();

            foreach (string u in callbacks.Keys)
            {
                IAPlanCallback cb = callbacks[u];
                //send info about insertion of a new assignment
                cb.OnAssignmentUpdate(ad);
            }
        }

        public void DistributeAssignment(AssignmentDistr a, string originUser)
        {
            foreach (string u in callbacks.Keys)
            {
                //distribute changes to all users except to the user who made the change
                if (u != originUser)
                {
                    IAPlanCallback cb = callbacks[u];
                    cb.OnAssignmentUpdate(a);
                }
            }
        }

        public void UpdateAssignment(Assignment aInp)
        {
            DataClassesDataContext context = new DataClassesDataContext();

            var _ass = (from a in context.Assignments
                        where a.Id == aInp.Id
                        select a).Single();
            if (_ass.Employee != aInp.Employee || _ass.FromDate != aInp.FromDate ||
               _ass.ToDate != aInp.ToDate || _ass.IBase != aInp.IBase || _ass.State != aInp.State)
            {
                IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
                string user = client.GetUserId();
                assignmentAction(AssignmentAction.Update, user, DateTime.Now, _ass);
                
                updateUserAction(user);
                _ass.Employee = aInp.Employee;
                _ass.fix = aInp.fix;
                _ass.FromDate = aInp.FromDate;
                _ass.ToDate = aInp.ToDate;
                _ass.IBase = aInp.IBase;
                _ass.Type = aInp.Type;
                _ass.State = aInp.State;

                context.SubmitChanges();

                foreach (string u in callbacks.Keys)
                {
                    IAPlanCallback cb = callbacks[u];
                    //send info about update of an assignment (with the old object)
                }
            }
        }
        #endregion

        public void InsertMessage(Message msg)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            context.Messages.InsertOnSubmit(msg);
            context.SubmitChanges();

            distributeMessages();
        }

        public void UpdateMessage(Message msg)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            var m = (from mes in context.Messages
                     where mes.id == msg.id
                     select mes).Single();
            m.Messagetext = msg.Messagetext;
            m.ValidTo = msg.ValidTo;
            context.SubmitChanges();

            distributeMessages();
        }

        public void DeleteMessage(Message msg)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            var m = (from mes in context.Messages
                     where mes.id == msg.id
                     select mes).Single();
            context.Messages.DeleteOnSubmit(m);
            context.SubmitChanges();

            distributeMessages();
        }

        private void distributeMessages()
        {
            DataClassesDataContext context = new DataClassesDataContext();

            var msg = (from m in context.Messages orderby m.PostingDate descending select m).ToArray();
            foreach (IAPlanCallback c in callbacks.Values)
            {
                c.OnMessageUpdate(msg);
            }
        }

        public void UndoRedo(Boolean undo)
        {
            DataClassesDataContext context = new DataClassesDataContext();
            IAPlanCallback client = OperationContext.Current.GetCallbackChannel<IAPlanCallback>();
            string user = client.GetUserId();
            DateTime maxdat;
            if(undo)
                maxdat = (from a in context.AssignmentUndoRedos
                          where a.ActionUser == user && a.Undo == undo
                          select a.ActionDate).Max();
            else
                maxdat = (from a in context.AssignmentUndoRedos
                          where a.ActionUser == user && a.Undo == undo
                          select a.ActionDate).Min();
            var actions = from act in context.AssignmentUndoRedos
                          where act.ActionUser == user
                           && act.ActionDate == maxdat
                           && act.Undo == undo
                          select act;
            
            foreach(AssignmentUndoRedo u in actions)
            {
                if (u.Action == (int)AssignmentAction.Update)
                {
                    var assign = (from a in context.Assignments
                                  where a.Id == u.Id select a).Single();
                    
                    var oldval = new Assignment();
                    oldval.Employee = assign.Employee;
                    oldval.fix = assign.fix;
                    oldval.FromDate = assign.FromDate;
                    oldval.IBase = assign.IBase;
                    oldval.State = assign.State;
                    oldval.ToDate = assign.ToDate;
                    oldval.Type = assign.Type;

                    assign.Employee = u.Emplyoee;
                    assign.fix = u.Fix;
                    assign.FromDate = u.FromDate;
                    assign.IBase = u.IBase;
                    assign.State = u.State;
                    assign.ToDate = u.ToDate;
                    assign.Type = u.Type;

                    foreach (string usr in callbacks.Keys)
                    {
                        IAPlanCallback cb = callbacks[usr];
                        cb.OnUndoRedo(u);
                    }

                    // make an undo from a redo an contrary
                    // save the old values
                    u.Undo = !undo;                    
                    u.Emplyoee = oldval.Employee;
                    u.Fix = oldval.fix;
                    u.FromDate = oldval.FromDate;
                    u.IBase = oldval.IBase;
                    u.State = oldval.State;
                    u.ToDate = oldval.ToDate;
                    u.Type = oldval.Type;                                        
                }
                if (u.Action == (int)AssignmentAction.Insert || u.Action == (int)AssignmentAction.Delete)
                {                    
                    if ((u.Undo && u.Action == (int)AssignmentAction.Insert) || (!u.Undo && u.Action == (int)AssignmentAction.Delete))
                    {
                        var assignment = (from a in context.Assignments
                                          where a.Id == u.Id
                                          select a).Single();
                        context.Assignments.DeleteOnSubmit(assignment);
                        foreach (string usr in callbacks.Keys)
                        {
                            IAPlanCallback cb = callbacks[usr];
                            cb.OnUndoRedo(u);
                        }
                    }
                    else
                    {
                        Assignment assignment = new Assignment();
                        // this works only for single user enviroments
                        // genereate a new id for productiy multi user environments
                        assignment.Id = u.Id;
                        assignment.Employee = u.Emplyoee;
                        assignment.fix = u.Fix;
                        assignment.FromDate = u.FromDate;
                        assignment.IBase = u.IBase;
                        assignment.State = u.State;
                        assignment.ToDate = u.ToDate;
                        assignment.Type = u.Type;
                        context.Assignments.InsertOnSubmit(assignment);
                        context.SubmitChanges();
                        var aib = (from av in context.AssignmentIBaseViews
                                  where av.Id == u.Id
                                  select av).Single();
                        AssignmentDistr ad = new AssignmentDistr();
                        ad.DBAssignment = assignment;
                        ad.AssignIBaseView = aib;
                        foreach (string usr in callbacks.Keys)
                        {
                            IAPlanCallback cb = callbacks[usr];
                            //cb.OnUndoRedo(u);
                            cb.OnAssignmentUpdate(ad);
                        }
                    }
                    
                    u.Undo = !undo;                    
                }
            }
            
            context.SubmitChanges();
            setUndoRedoButtons(user);
        }

    }
}
